@@ -187,7 +187,10 @@ readRegStr (const char *root, const char *dir, const char *name)
         return ret;
     }
     ret = _readRegStr (root_key, dir, name, false);
-
+    if (ret.empty()) {
+        // Try alternate as fallback
+        ret = _readRegStr (root_key, dir, name, true);
+    }
     if (ret.empty()) {
         // Try local machine as fallback.
         ret = _readRegStr (HKEY_LOCAL_MACHINE, dir, name, false);
@@ -461,7 +464,7 @@ getTmpPathUtf8 ()
    Returns the actually used file name in case there
    were other files with that name. */
 wchar_t*
-get_tmp_outfile (wchar_t *name, HANDLE *outHandle)
+get_tmp_outfile (const wchar_t *name, HANDLE *outHandle)
 {
   const auto utf8Name = wchar_to_utf8_string (name);
   const auto tmpPath = getTmpPathUtf8 ();
