@@ -37,6 +37,7 @@
 #include "cpphelp.h"
 #include "gpgoladdin.h"
 #include "categorymanager.h"
+#include "recipient.h"
 
 HRESULT
 gpgol_queryInterface (LPUNKNOWN pObj, REFIID riid, LPVOID FAR *ppvObj)
@@ -98,7 +99,7 @@ get_object_name (LPUNKNOWN obj)
   disp->GetTypeInfo (0, 0, &tinfo);
   if (!tinfo)
     {
-      log_debug ("%s:%s: no typeinfo found for object\n", 
+      log_debug ("%s:%s: no typeinfo found for object\n",
                  SRCNAME, __func__);
       goto leave;
     }
@@ -106,7 +107,7 @@ get_object_name (LPUNKNOWN obj)
   bstrname = NULL;
   hr = tinfo->GetDocumentation (MEMBERID_NIL, &bstrname, 0, 0, 0);
   if (hr || !bstrname)
-    log_debug ("%s:%s: GetDocumentation failed: hr=%#lx\n", 
+    log_debug ("%s:%s: GetDocumentation failed: hr=%#lx\n",
                SRCNAME, __func__, hr);
   if (bstrname)
     {
@@ -123,6 +124,24 @@ get_object_name (LPUNKNOWN obj)
   TRETURN name;
 }
 
+std::string
+get_object_name_s (LPUNKNOWN obj)
+{
+  char *name = get_object_name (obj);
+  std::string ret = "(null)";
+  if (name)
+    {
+      ret = name;
+    }
+  xfree (name);
+  return ret;
+}
+
+std::string
+get_object_name_s (shared_disp_t obj)
+{
+  return get_object_name_s (obj.get ());
+}
 
 /* Lookup the dispid of object PDISP for member NAME.  Returns
    DISPID_UNKNOWN on error.  */
@@ -144,7 +163,7 @@ lookup_oom_dispid (LPDISPATCH pDisp, const char *name)
       TRETURN DISPID_UNKNOWN;/* Error:  Out of memory.  */
     }
 
-  hr = pDisp->GetIDsOfNames (IID_NULL, &wname, 1, 
+  hr = pDisp->GetIDsOfNames (IID_NULL, &wname, 1,
                              LOCALE_SYSTEM_DEFAULT, &dispid);
   xfree (wname);
   if (hr != S_OK || dispid == DISPID_UNKNOWN)
@@ -263,12 +282,12 @@ get_oom_object (LPDISPATCH pStart, const char *fullname)
             }
           TRETURN pDisp; /* Ready.  */
         }
-      
+
       /* Break out the next name part.  */
       {
         const char *dot;
         size_t n;
-        
+
         dot = strchr (fullname, '.');
         if (dot == fullname)
           {
@@ -279,18 +298,18 @@ get_oom_object (LPDISPATCH pStart, const char *fullname)
           n = dot - fullname;
         else
           n = strlen (fullname);
-        
+
         if (n >= sizeof name)
           n = sizeof name - 1;
         strncpy (name, fullname, n);
         name[n] = 0;
-        
+
         if (dot)
           fullname = dot + 1;
         else
           fullname += strlen (fullname);
       }
-      
+
       if (!strncmp (name, "get_", 4) && name[4])
         {
           dispmethod = DISPATCH_PROPERTYGET;
@@ -347,14 +366,14 @@ get_oom_object (LPDISPATCH pStart, const char *fullname)
         {
           if (n_parms == 4)
             {
-              dispparams.rgvarg[0].vt = VT_ERROR; 
-              dispparams.rgvarg[0].scode = DISP_E_PARAMNOTFOUND; 
-              dispparams.rgvarg[1].vt = VT_ERROR; 
-              dispparams.rgvarg[1].scode = DISP_E_PARAMNOTFOUND; 
-              dispparams.rgvarg[2].vt = VT_INT; 
-              dispparams.rgvarg[2].intVal = parmint; 
-              dispparams.rgvarg[3].vt = VT_ERROR; 
-              dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND; 
+              dispparams.rgvarg[0].vt = VT_ERROR;
+              dispparams.rgvarg[0].scode = DISP_E_PARAMNOTFOUND;
+              dispparams.rgvarg[1].vt = VT_ERROR;
+              dispparams.rgvarg[1].scode = DISP_E_PARAMNOTFOUND;
+              dispparams.rgvarg[2].vt = VT_INT;
+              dispparams.rgvarg[2].intVal = parmint;
+              dispparams.rgvarg[3].vt = VT_ERROR;
+              dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND;
               dispparams.cArgs = n_parms;
             }
           else if (n_parms == 1 && parmstr)
@@ -391,6 +410,17 @@ get_oom_object (LPDISPATCH pStart, const char *fullname)
   TRETURN NULL;
 }
 
+shared_disp_t
+get_oom_object_s (shared_disp_t pStart, const char *fullname)
+{
+  return MAKE_SHARED (get_oom_object (pStart.get (), fullname));
+}
+
+shared_disp_t
+get_oom_object_s (LPDISPATCH pStart, const char *fullname)
+{
+  return MAKE_SHARED (get_oom_object (pStart, fullname));
+}
 
 /* Helper for put_oom_icon.  */
 static int
@@ -412,7 +442,7 @@ put_picture_or_mask (LPDISPATCH pDisp, int resource, int size, int is_mask)
   fuload = LR_CREATEDIBSECTION | LR_SHARED;
   if (is_mask)
     fuload |= LR_MONOCHROME;
-  
+
   memset (&pdesc, 0, sizeof pdesc);
   pdesc.cbSizeofstruct = sizeof pdesc;
   pdesc.picType = PICTYPE_BITMAP;
@@ -421,13 +451,13 @@ put_picture_or_mask (LPDISPATCH pDisp, int resource, int size, int is_mask)
                                            IMAGE_BITMAP, size, size, fuload);
   if (!pdesc.bmp.hbitmap)
     {
-      log_error_w32 (-1, "%s:%s: LoadImage(%d) failed\n", 
+      log_error_w32 (-1, "%s:%s: LoadImage(%d) failed\n",
                      SRCNAME, __func__, resource);
       TRETURN -1;
     }
 
   /* Wrap the image into an OLE object.  */
-  hr = OleCreatePictureIndirect (&pdesc, IID_IPictureDisp, 
+  hr = OleCreatePictureIndirect (&pdesc, IID_IPictureDisp,
                                  TRUE, (void **) &pPict);
   if (hr != S_OK || !pPict)
     {
@@ -435,7 +465,7 @@ put_picture_or_mask (LPDISPATCH pDisp, int resource, int size, int is_mask)
                  SRCNAME, __func__, hr);
       TRETURN -1;
     }
-        
+
   /* Store to the Picture or Mask property of the CommandBarButton.  */
   dispid = lookup_oom_dispid (pDisp, is_mask? "Mask":"Picture");
 
@@ -507,7 +537,7 @@ put_oom_bool (LPDISPATCH pDisp, const char *name, int value)
                       NULL, NULL, NULL);
   if (hr != S_OK)
     {
-      log_debug ("%s:%s: Putting '%s' failed: %#lx", 
+      log_debug ("%s:%s: Putting '%s' failed: %#lx",
                  SRCNAME, __func__, name, hr);
       TRETURN -1;
     }
@@ -543,13 +573,77 @@ put_oom_int (LPDISPATCH pDisp, const char *name, int value)
                       NULL, NULL, NULL);
   if (hr != S_OK)
     {
-      log_debug ("%s:%s: Putting '%s' failed: %#lx", 
+      log_debug ("%s:%s: Putting '%s' failed: %#lx",
                  SRCNAME, __func__, name, hr);
       TRETURN -1;
     }
   TRETURN 0;
 }
 
+/* Set the property NAME to VALUE.  */
+int
+put_oom_array (LPDISPATCH pDisp, const char *name, unsigned char *value,
+               size_t size)
+{
+  TSTART;
+  HRESULT hr;
+  DISPID dispid_put = DISPID_PROPERTYPUT;
+  DISPID dispid;
+  DISPPARAMS dispparams;
+  VARIANT aVariant[1];
+  unsigned int argErr = 0;
+  EXCEPINFO execpinfo;
+  init_excepinfo (&execpinfo);
+
+  VariantInit (aVariant);
+
+  dispid = lookup_oom_dispid (pDisp, name);
+  if (dispid == DISPID_UNKNOWN)
+    {
+      TRETURN -1;
+    }
+  /* Prepare the savearray */
+  SAFEARRAYBOUND saBound;
+  saBound.lLbound = 0;
+  saBound.cElements = size;
+  SAFEARRAY* psa = SafeArrayCreate(VT_UI1, 1, &saBound);
+
+  if (!psa)
+    {
+      log_err ("Failed to create SafeArray");
+      TRETURN -1;
+    }
+
+  hr = SafeArrayLock(psa);
+  if (!SUCCEEDED (hr))
+    {
+      log_err ("Failed to lock array.");
+    }
+  memcpy (psa->pvData, value, size);
+  SafeArrayUnlock(psa);
+
+  dispparams.rgvarg = aVariant;
+  dispparams.rgvarg[0].vt = VT_ARRAY;
+  dispparams.rgvarg[0].parray = psa;
+  dispparams.cArgs = 1;
+  dispparams.rgdispidNamedArgs = &dispid_put;
+  dispparams.cNamedArgs = 1;
+  hr = pDisp->Invoke (dispid, IID_NULL, LOCALE_SYSTEM_DEFAULT,
+                      DISPATCH_PROPERTYPUT, &dispparams,
+                      NULL, &execpinfo, &argErr);
+  SafeArrayDestroy(psa);
+  if (hr != S_OK)
+    {
+      log_debug ("%s:%s: error: putting %s"
+                 " hr=0x%x argErr=0x%x",
+                 SRCNAME, __func__, name,
+                 (unsigned int)hr,
+                 (unsigned int)argErr);
+      dump_excepinfo (execpinfo);
+      TRETURN -1;
+    }
+  TRETURN 0;
+}
 
 /* Set the property NAME to STRING.  */
 int
@@ -594,7 +688,7 @@ put_oom_string (LPDISPATCH pDisp, const char *name, const char *string)
   SysFreeString (bstring);
   if (hr != S_OK)
     {
-      log_debug ("%s:%s: Putting '%s' failed: %#lx", 
+      log_debug ("%s:%s: Putting '%s' failed: %#lx",
                  SRCNAME, __func__, name, hr);
       dump_excepinfo (execpinfo);
       TRETURN -1;
@@ -646,10 +740,10 @@ int
 get_oom_bool (LPDISPATCH pDisp, const char *name)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   int result = 0;
   DISPID dispid;
-  
+
   dispid = lookup_oom_dispid (pDisp, name);
   if (dispid != DISPID_UNKNOWN)
     {
@@ -681,10 +775,10 @@ int
 get_oom_int (LPDISPATCH pDisp, const char *name)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   int result = 0;
   DISPID dispid;
-  
+
   dispid = lookup_oom_dispid (pDisp, name);
   if (dispid != DISPID_UNKNOWN)
     {
@@ -709,6 +803,78 @@ get_oom_int (LPDISPATCH pDisp, const char *name)
   TRETURN result;
 }
 
+int
+get_oom_int (shared_disp_t pDisp, const char *name)
+{
+  return get_oom_int (pDisp.get (), name);
+}
+
+int
+get_oom_dirty (LPDISPATCH pDisp)
+{
+  TSTART;
+  HRESULT hr;
+  DISPID dispid = DISPID_DIRTY_RAT;
+
+  DISPPARAMS dispparams = {NULL, NULL, 0, 0};
+  VARIANT rVariant;
+
+  VariantInit (&rVariant);
+  hr = pDisp->Invoke (dispid, IID_NULL, LOCALE_SYSTEM_DEFAULT,
+                      DISPATCH_PROPERTYGET | DISPATCH_METHOD,
+                      &dispparams, &rVariant, NULL, NULL);
+  if (hr != S_OK)
+    {
+      log_debug ("%s:%s: Property dirty not found: %#lx",
+                 SRCNAME, __func__, hr);
+      TRETURN -1;
+    }
+  return !!rVariant.bVal;
+}
+
+
+#if 0
+int
+put_oom_dirty (LPDISPATCH pDisp, bool value)
+{
+  TSTART;
+
+  /* NOTE: I have found no scenario where this does
+     not return the exception that the property is
+     write protected. But we can never know when
+     we need such an arcane function so I left it in. */
+
+  HRESULT hr;
+  DISPID dispid_put = DISPID_PROPERTYPUT;
+  DISPID dispid = DISPID_DIRTY_RAT;
+  DISPPARAMS dispparams;
+  VARIANT aVariant[1];
+  unsigned int argErr = 0;
+  EXCEPINFO execpinfo;
+  init_excepinfo (&execpinfo);
+
+  dispparams.rgvarg = aVariant;
+  dispparams.rgvarg[0].vt = VT_BOOL;
+  dispparams.rgvarg[0].boolVal = value? VARIANT_TRUE:VARIANT_FALSE;
+  dispparams.cArgs = 1;
+  dispparams.rgdispidNamedArgs = &dispid_put;
+  dispparams.cNamedArgs = 1;
+  hr = pDisp->Invoke (dispid, IID_NULL, LOCALE_SYSTEM_DEFAULT,
+                      DISPATCH_PROPERTYPUT | DISPATCH_METHOD, &dispparams,
+                      NULL, &execpinfo, &argErr);
+  if (hr != S_OK)
+    {
+      log_debug ("%s:%s: error: invoking dirty p=%p vt=%d"
+                 " hr=0x%x argErr=0x%x",
+                 SRCNAME, __func__,
+                 nullptr, 0, (unsigned int)hr,
+                 (unsigned int)argErr);
+      dump_excepinfo (execpinfo);
+      TRETURN -1;
+    }
+  TRETURN 0;
+}
+#endif
 
 /* Get the string property NAME of the object PDISP.  Returns NULL if
    not found or if it is not a string property.  */
@@ -716,10 +882,10 @@ char *
 get_oom_string (LPDISPATCH pDisp, const char *name)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   char *result = NULL;
   DISPID dispid;
-  
+
   dispid = lookup_oom_dispid (pDisp, name);
   if (dispid != DISPID_UNKNOWN)
     {
@@ -744,16 +910,33 @@ get_oom_string (LPDISPATCH pDisp, const char *name)
   TRETURN result;
 }
 
+std::string
+get_oom_string_s (LPDISPATCH pDisp, const char *name)
+{
+  char *ret_c =  get_oom_string (pDisp, name);
+  std::string ret;
+  if (ret_c)
+    {
+       ret = ret_c;
+       xfree (ret_c);
+    }
+  return ret;
+}
 
+std::string
+get_oom_string_s (shared_disp_t pDisp, const char *name)
+{
+  return get_oom_string_s (pDisp.get (), name);
+}
 /* Get the object property NAME of the object PDISP.  Returns NULL if
    not found or if it is not an object perty.  */
 LPUNKNOWN
 get_oom_iunknown (LPDISPATCH pDisp, const char *name)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   DISPID dispid;
-  
+
   dispid = lookup_oom_dispid (pDisp, name);
   if (dispid != DISPID_UNKNOWN)
     {
@@ -790,7 +973,7 @@ LPDISPATCH
 get_oom_control_bytag (LPDISPATCH pDisp, const char *tag)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   DISPID dispid;
   DISPPARAMS dispparams;
   VARIANT aVariant[4];
@@ -818,13 +1001,13 @@ get_oom_control_bytag (LPDISPATCH pDisp, const char *tag)
   }
   dispparams.rgvarg = aVariant;
   dispparams.rgvarg[0].vt = VT_ERROR; /* Visible */
-  dispparams.rgvarg[0].scode = DISP_E_PARAMNOTFOUND; 
+  dispparams.rgvarg[0].scode = DISP_E_PARAMNOTFOUND;
   dispparams.rgvarg[1].vt = VT_BSTR;  /* Tag */
   dispparams.rgvarg[1].bstrVal = bstring;
   dispparams.rgvarg[2].vt = VT_ERROR; /* Id */
   dispparams.rgvarg[2].scode = DISP_E_PARAMNOTFOUND;
   dispparams.rgvarg[3].vt = VT_ERROR;/* Type */
-  dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND; 
+  dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND;
   dispparams.cArgs = 4;
   dispparams.cNamedArgs = 0;
   VariantInit (&rVariant);
@@ -858,7 +1041,7 @@ LPDISPATCH
 add_oom_button (LPDISPATCH pObj)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   DISPID dispid;
   DISPPARAMS dispparams;
   VARIANT aVariant[5];
@@ -870,11 +1053,11 @@ add_oom_button (LPDISPATCH pObj)
   dispparams.rgvarg[0].vt = VT_BOOL;  /* Temporary */
   dispparams.rgvarg[0].boolVal = VARIANT_TRUE;
   dispparams.rgvarg[1].vt = VT_ERROR;  /* Before */
-  dispparams.rgvarg[1].scode = DISP_E_PARAMNOTFOUND; 
+  dispparams.rgvarg[1].scode = DISP_E_PARAMNOTFOUND;
   dispparams.rgvarg[2].vt = VT_ERROR;  /* Parameter */
-  dispparams.rgvarg[2].scode = DISP_E_PARAMNOTFOUND; 
+  dispparams.rgvarg[2].scode = DISP_E_PARAMNOTFOUND;
   dispparams.rgvarg[3].vt = VT_ERROR;  /* Id */
-  dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND; 
+  dispparams.rgvarg[3].scode = DISP_E_PARAMNOTFOUND;
   dispparams.rgvarg[4].vt = VT_INT;    /* Type */
   dispparams.rgvarg[4].intVal = MSOCONTROLBUTTON;
   dispparams.cArgs = 5;
@@ -900,7 +1083,7 @@ void
 del_oom_button (LPDISPATCH pObj)
 {
   TSTART;
-  HRESULT hr;      
+  HRESULT hr;
   DISPID dispid;
   DISPPARAMS dispparams;
   VARIANT aVariant[5];
@@ -1044,8 +1227,8 @@ put_pa_int (LPDISPATCH pDisp, const char *dasl_id, int value)
   TSTART;
   VARIANT var;
   VariantInit (&var);
-  var.vt = VT_INT;
-  var.intVal = value;
+  var.vt = VT_I4;
+  var.lVal = value;
   int ret = put_pa_variant (pDisp, dasl_id, &var);
   VariantClear (&var);
   TRETURN ret;
@@ -1276,7 +1459,8 @@ get_recipient_addr_fallbacks (LPDISPATCH recipient)
 */
 static bool
 try_resolve_group (LPDISPATCH addrEntry,
-                   std::vector<std::pair<std::string, shared_disp_t> > &ret)
+                   std::vector<std::pair<Recipient, shared_disp_t> >&ret,
+                   int recipient_type)
 {
   TSTART;
   /* Get the name for debugging */
@@ -1288,12 +1472,12 @@ try_resolve_group (LPDISPATCH addrEntry,
     }
   xfree (cname);
 
-  int type = get_oom_int (addrEntry, "AddressEntryUserType");
+  int user_type = get_oom_int (addrEntry, "AddressEntryUserType");
 
-  if (type != DISTRIBUTION_LIST_ADDRESS_ENTRY_TYPE)
+  if (user_type != DISTRIBUTION_LIST_ADDRESS_ENTRY_TYPE)
     {
       log_data ("%s:%s: type of %s is %i",
-                       SRCNAME, __func__, anonstr (name.c_str()), type);
+                       SRCNAME, __func__, anonstr (name.c_str()), user_type);
       TRETURN false;
     }
 
@@ -1340,14 +1524,14 @@ try_resolve_group (LPDISPATCH addrEntry,
           log_debug ("%s:%s: recursive address entry %s",
                      SRCNAME, __func__,
                      anonstr (entryName.c_str()));
-          if (try_resolve_group (entry.get(), ret))
+          if (try_resolve_group (entry.get(), ret, recipient_type))
             {
               foundOne = true;
               continue;
             }
         }
 
-      std::pair<std::string, shared_disp_t> element;
+      std::pair<Recipient, shared_disp_t> element;
       element.second = entry;
 
       /* Resolve directly ? */
@@ -1358,7 +1542,8 @@ try_resolve_group (LPDISPATCH addrEntry,
           char *resolved = get_pa_string (entry.get(), PR_EMAIL_ADDRESS_DASL);
           if (resolved)
             {
-              element.first = resolved;
+              element.first = Recipient (resolved, entryName.c_str (),
+                                         recipient_type);
               ret.push_back (element);
               foundOne = true;
               continue;
@@ -1370,7 +1555,8 @@ try_resolve_group (LPDISPATCH addrEntry,
       char *ex_resolved = get_recipient_addr_entry_fallbacks_ex (entry.get());
       if (ex_resolved)
         {
-          element.first = ex_resolved;
+          element.first = Recipient (ex_resolved, entryName.c_str (),
+                                     recipient_type);
           ret.push_back (element);
           foundOne = true;
           continue;
@@ -1392,12 +1578,12 @@ try_resolve_group (LPDISPATCH addrEntry,
 
 /* Get the recipient mbox addresses with the addrEntry
    object corresponding to the resolved address. */
-std::vector<std::pair<std::string, shared_disp_t> >
+std::vector<std::pair<Recipient, shared_disp_t> >
 get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
 {
   TSTART;
   int recipientsCnt = get_oom_int (recipients, "Count");
-  std::vector<std::pair<std::string, shared_disp_t> > ret;
+  std::vector<std::pair<Recipient, shared_disp_t> > ret;
   int i;
 
   if (!recipientsCnt)
@@ -1424,8 +1610,19 @@ get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
           break;
         }
 
+      int recipient_type = get_oom_int (recipient, "Type");
+      std::string entryName;
+      char *entry_name = get_oom_string (recipient, "Name");
+      if (entry_name)
+        {
+          entryName = entry_name;
+          xfree (entry_name);
+        }
+
+
       auto addrEntry = MAKE_SHARED (get_oom_object (recipient, "AddressEntry"));
-      if (addrEntry && try_resolve_group (addrEntry.get (), ret))
+      if (addrEntry && try_resolve_group (addrEntry.get (), ret,
+                                          recipient_type))
         {
           log_debug ("%s:%s: Resolved recipient group",
                      SRCNAME, __func__);
@@ -1433,13 +1630,15 @@ get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
           continue;
         }
 
-      std::pair<std::string, shared_disp_t> entry;
+      std::pair<Recipient, shared_disp_t> entry;
       entry.second = addrEntry;
 
       char *resolved = get_pa_string (recipient, PR_SMTP_ADDRESS_DASL);
       if (resolved)
         {
-          entry.first = resolved;
+          entry.first = Recipient (resolved, entryName.c_str (),
+                                   recipient_type);
+          entry.first.setIndex (i);
           xfree (resolved);
           gpgol_release (recipient);
           ret.push_back (entry);
@@ -1449,7 +1648,9 @@ get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
       resolved = get_recipient_addr_fallbacks (recipient);
       if (resolved)
         {
-          entry.first = resolved;
+          entry.first = Recipient (resolved, entryName.c_str (),
+                                   recipient_type);
+          entry.first.setIndex (i);
           xfree (resolved);
           gpgol_release (recipient);
           ret.push_back (entry);
@@ -1463,7 +1664,8 @@ get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
                  SRCNAME, __func__);
       if (address)
         {
-          entry.first = address;
+          entry.first = Recipient (resolved, recipient_type);
+          entry.first.setIndex (i);
           ret.push_back (entry);
           xfree (address);
         }
@@ -1476,11 +1678,11 @@ get_oom_recipients_with_addrEntry (LPDISPATCH recipients, bool *r_err)
 }
 
 /* Gets the resolved smtp addresses of the recpients. */
-std::vector<std::string>
+std::vector<Recipient>
 get_oom_recipients (LPDISPATCH recipients, bool *r_err)
 {
   TSTART;
-  std::vector<std::string> ret;
+  std::vector<Recipient> ret;
   for (const auto pair: get_oom_recipients_with_addrEntry (recipients, r_err))
     {
       ret.push_back (pair.first);
@@ -3120,16 +3322,304 @@ count_visible_attachments (LPDISPATCH attachments)
         }
       VARIANT var;
       VariantInit (&var);
-      if (get_pa_variant (oom_attach, PR_ATTACHMENT_HIDDEN_DASL, &var) ||
-          (var.vt == VT_BOOL && var.boolVal == VARIANT_FALSE))
+      if (get_pa_variant (oom_attach, PR_ATTACHMENT_HIDDEN_DASL, &var))
+        {
+          /* SECURITY: Testing has shown that for all mail types GpgOL
+             handles that might contain an unsigned attachment we always
+             have the MAPIOBJECT / get the hidden state. Only the transient
+             MIME attachments. The ones used for the MAPI to MIME conversion
+             and which are hidden by GpgOL and handled by GpgOL will have
+             no MAPIOBJECT when a mail is opened from file. So this will
+             remove the warning that "smime.p7m" or "gpgol_mime_structure.txt"
+             are unsigned and unencrypted attachments. */
+          log_dbg ("Failed to get hidden state.");
+          LPUNKNOWN mapiobj = get_oom_iunknown (oom_attach, "MAPIOBJECT");
+          if (!mapiobj)
+            {
+              const auto dispName = get_oom_string_s (oom_attach,
+                                                      "DisplayName");
+              log_dbg ("Attachment: %s has no mapiobject. Ignoring it.",
+                       anonstr (dispName.c_str ()));
+            }
+          else
+            {
+              gpgol_release (mapiobj);
+              const auto dispName = get_oom_string_s (oom_attach,
+                                                      "DisplayName");
+              log_dbg ("Attachment %s without hidden state but mapiobj. "
+                       "Count as visible.", anonstr (dispName.c_str ()));
+              ret++;
+            }
+        }
+      else if (var.vt == VT_BOOL && var.boolVal == VARIANT_FALSE)
         {
            ret++;
         }
-      else
-        {
-          gpgol_release (oom_attach);
-        }
+      gpgol_release (oom_attach);
       VariantClear (&var);
     }
   return ret;
 }
+
+int invoke_oom_method_with_int (LPDISPATCH pDisp, const char *name,
+                                int arg,
+                                VARIANT *rVariant)
+{
+  TSTART;
+  DISPPARAMS parms;
+  VARIANT argvars[1];
+  VariantInit (&argvars[0]);
+  argvars[0].vt = VT_INT;
+  argvars[0].intVal = arg;
+  parms.cArgs = 1;
+  parms.cNamedArgs = 0;
+  parms.rgvarg = argvars;
+
+  TRETURN invoke_oom_method_with_parms (pDisp, name,
+                                        rVariant, &parms);
+
+}
+
+int invoke_oom_method_with_string (LPDISPATCH pDisp, const char *name,
+                                   const char *arg,
+                                   VARIANT *rVariant)
+{
+  TSTART;
+  if (!arg)
+    {
+      TRETURN 0;
+    }
+  wchar_t *warg = utf8_to_wchar (arg);
+  if (!warg)
+    {
+      TRETURN 1;
+    }
+  VARIANT aVariant[1];
+  VariantInit (&aVariant[0]);
+  aVariant[0].vt = VT_BSTR;
+  aVariant[0].bstrVal = SysAllocString (warg);
+  xfree (warg);
+
+  DISPPARAMS dispparams;
+  dispparams.rgvarg = aVariant;
+  dispparams.cArgs = 1;
+  dispparams.cNamedArgs = 0;
+
+  int ret = invoke_oom_method_with_parms (pDisp, name, rVariant, &dispparams);
+  VariantClear(&aVariant[0]);
+  TRETURN ret;
+}
+
+int
+set_oom_recipients (LPDISPATCH item, const std::vector<Recipient> &recps)
+{
+  if (!item)
+    {
+      STRANGEPOINT;
+      TRETURN -1;
+    }
+
+  auto oom_recps = MAKE_SHARED (get_oom_object (item, "Recipients"));
+
+  if (!oom_recps)
+    {
+      STRANGEPOINT;
+      TRETURN -1;
+    }
+
+  /* First clear out the current recipients. */
+  int ret = invoke_oom_method (oom_recps.get (), "RemoveAll", nullptr);
+
+  if (ret)
+    {
+      STRANGEPOINT;
+      TRETURN ret;
+    }
+
+  for (const auto &recp: recps)
+    {
+      if (recp.type() == Recipient::olOriginator)
+        {
+          /* Skip the originator, we only add it internally but
+             it does not need to be in OOM. */
+          continue;
+        }
+      VARIANT result;
+      VariantInit (&result);
+      ret = invoke_oom_method_with_string (oom_recps.get (), "Add",
+                                           recp.mbox ().c_str (),
+                                           &result);
+      if (ret)
+        {
+          log_err ("Failed to add recipient.");
+          TRETURN ret;
+        }
+      if (result.vt != VT_DISPATCH || !result.pdispVal)
+        {
+          log_err ("No recipient result.");
+          continue;
+        }
+      if (put_oom_int (result.pdispVal, "Type", recp.type()))
+        {
+          log_err ("Failed to set recipient type.");
+        }
+      /* This releases the recipient. */
+      VariantClear (&result);
+    }
+  TRETURN 0;
+}
+
+int
+remove_oom_recipient (LPDISPATCH item, const std::string &mbox)
+{
+  TSTART;
+  if (!item)
+    {
+      STRANGEPOINT;
+      TRETURN -1;
+    }
+
+  auto oom_recps = MAKE_SHARED (get_oom_object (item, "Recipients"));
+
+  if (!oom_recps)
+    {
+      STRANGEPOINT;
+      TRETURN -1;
+    }
+
+  bool r_err = false;
+  const auto recps = get_oom_recipients (oom_recps.get (), &r_err);
+  if (r_err)
+    {
+      log_debug ("Failure to lookup recipients via OOM");
+      TRETURN -1;
+    }
+
+  for (const auto &recp: recps)
+    {
+      if (recp.mbox () == mbox && recp.index () != -1)
+        {
+          TRETURN invoke_oom_method_with_int (oom_recps.get (),
+                                              "Remove", recp.index (),
+                                              nullptr);
+        }
+    }
+  TRETURN -1;
+}
+
+void
+oom_dump_idispatch (LPDISPATCH obj)
+{
+  log_dbg ("Start infos about %p", obj);
+  if (!obj)
+    {
+      log_dbg ("It's NULL");
+      return;
+    }
+  log_dbg ("Name: '%s'", get_object_name_s (obj).c_str ());
+
+  LPTYPEINFO typeinfo = nullptr;
+  HRESULT hr = obj->GetTypeInfo (0, 0, &typeinfo);
+
+  if (!typeinfo || FAILED (hr))
+    {
+      log_dbg ("No typeinfo.");
+      return;
+    }
+
+  TYPEATTR* pta = NULL;
+  hr = typeinfo->GetTypeAttr(&pta);
+
+  if (!pta || FAILED (hr))
+    {
+      log_dbg ("No type attr");
+      return;
+    }
+
+  /* First the IID to have it clear */
+  LPOLESTR lpsz = NULL;
+  hr = StringFromIID(pta->guid, &lpsz);
+  if(FAILED (hr))
+    {
+      hr = StringFromCLSID (pta->guid, &lpsz);
+    }
+
+  if(SUCCEEDED (hr))
+    {
+      log_dbg ("Interface: %S", lpsz);
+      CoTaskMemFree(lpsz);
+    }
+
+  FUNCDESC *pfd = nullptr;
+  /* Lets see what functions we have. */
+  for(int i = 0; i < pta->cFuncs; i++)
+    {
+      typeinfo->GetFuncDesc(i, &pfd);
+
+      BSTR names[1];
+      unsigned int dumb;
+
+      typeinfo->GetNames(pfd->memid, names, 1, &dumb);
+      if (!names[0])
+        {
+          typeinfo->ReleaseFuncDesc(pfd);
+          continue;
+        }
+
+      log_dbg ("%i: %S id=0x%li With %d param(s)\n", i,
+               (names[0]), pfd->memid, pfd->cParams);
+      typeinfo->ReleaseFuncDesc(pfd);
+      SysFreeString(names[0]);
+    }
+  typeinfo->ReleaseTypeAttr(pta);
+
+  /* Now for some interesting object relations
+     that many oom objecs have. */
+  const char * relations[] = {
+    "Parent",
+    "GetInspector",
+    "Session",
+    "Sender",
+    nullptr
+  };
+
+  for (int i = 0; relations [i]; i++)
+    {
+      LPDISPATCH rel = get_oom_object (obj, relations[i]);
+      log_dbg ("%s: %s", relations [i],
+               get_object_name_s (rel).c_str ());
+      gpgol_release (rel);
+    }
+
+  /* Now for some interesting string values. */
+  const char * stringVals[] = {
+    "EntryID",
+    "Subject",
+    "MessageClass",
+    "Body",
+    nullptr
+  };
+
+  for (int i = 0; stringVals[i]; i++)
+    {
+      const auto str = get_oom_string_s (obj, stringVals[i]);
+      log_dbg ("%s: %s", stringVals[i], str.c_str ());
+    }
+
+  log_dbg ("Object dump done");
+  return;
+}
+
+int
+get_oom_crypto_flags (LPDISPATCH mailitem)
+{
+  TSTART;
+  int r_val = 0;
+  int err = get_pa_int (mailitem, PR_SECURITY_FLAGS_DASL, &r_val);
+  if (err)
+    {
+      log_dbg ("Failed to get security flags.");
+      TRETURN 0;
+    }
+  TRETURN r_val;
+}
