@@ -64,7 +64,7 @@ gpgol_window_proc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
                              SRCNAME, __func__);
                   TBREAK;
                 }
-              mail->parsing_done();
+              mail->parsingDone_o ();
               TBREAK;
             }
           case (RECIPIENT_ADDED):
@@ -102,7 +102,7 @@ gpgol_window_proc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
               invoke_oom_method (mail->item (), "Save", NULL);
               log_debug ("%s:%s: Revert mail. Save done. Updating body..",
                          SRCNAME, __func__);
-              mail->updateBody_o ();
+              mail->updateBody_o (false);
               log_debug ("%s:%s: Revert mail done.",
                          SRCNAME, __func__);
               TBREAK;
@@ -342,6 +342,7 @@ gpgol_window_proc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
                   xfree (data);
                   TBREAK;
                 }
+              memdbg_addRef (mapi_message);
               log_debug ("%s:%s: Restoring message class after move to: %s",
                          SRCNAME, __func__, data->old_class);
               mapi_set_mesage_class (mapi_message, data->old_class);
@@ -359,6 +360,50 @@ gpgol_window_proc (HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
               xfree (data);
               TBREAK;
             }
+          case SEND_MULTIPLE_MAILS:
+            {
+              auto mail = (Mail*) ctx->data;
+              if (!Mail::isValidPtr (mail))
+                {
+                  log_debug ("%s:%s: SEND_MULTIPLE_MAILS for mail which is gone.",
+                             SRCNAME, __func__);
+                  TBREAK;
+                }
+
+              log_debug ("%s:%s: Send multiple mails for %p.",
+                         SRCNAME, __func__, mail);
+              mail->splitAndSend_o ();
+              TBREAK;
+            }
+          case SEND:
+            {
+              auto mail = (Mail*) ctx->data;
+              if (!Mail::isValidPtr (mail))
+                {
+                  log_debug ("%s:%s: SEND for mail %p which is gone.",
+                             SRCNAME, __func__, mail);
+                  TBREAK;
+                }
+
+              log_debug ("%s:%s: Send for %p.",
+                         SRCNAME, __func__, mail);
+              invoke_oom_method (mail->item (), "Send", nullptr);
+              TBREAK;
+            }
+          case SHOW_PREVIEW:
+            {
+              auto mail = (Mail*) ctx->data;
+              if (!Mail::isValidPtr (mail))
+                {
+                  log_dbg ("SHOW_PREVIEW for mail %p which is gone.",
+                           mail);
+                  TBREAK;
+                }
+              log_dbg ("SHOW_PREVIEW for %p", mail);
+              mail->parsingDone_o (true);
+              TBREAK;
+            }
+
           default:
             log_debug ("%s:%s: Unknown msg %x",
                        SRCNAME, __func__, ctx->wmsg_type);
